# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tagging', '0003_auto_20170423_1143'),
    ]

    operations = [
        migrations.AddField(
            model_name='taggeditem',
            name='created_at',
            field=models.DateTimeField(null=True, db_index=True, auto_now_add=True),
        ),
        migrations.AddField(
            model_name='taggeditem',
            name='created_user',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



"""
Apps for tagging.
"""
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class TaggingConfig(AppConfig):
    """
    Config for Tagging application.
    """
    name = 'tagging'
    label = 'tagging'
    verbose_name = _('Tagging')



"""
Generic components for tagging.
"""
from django.contrib.contenttypes.models import ContentType


def fetch_content_objects(tagged_items, select_related_for=None):
    """
    Retrieves ``ContentType`` and content objects for the given list of
    ``TaggedItems``, grouping the retrieval of content objects by model
    type to reduce the number of queries executed.

    This results in ``number_of_content_types + 1`` queries rather than
    the ``number_of_tagged_items * 2`` queries you'd get by iterating
    over the list and accessing each item's ``object`` attribute.

    A ``select_related_for`` argument can be used to specify a list of
    of model names (corresponding to the ``model`` field of a
    ``ContentType``) for which ``select_related`` should be used when
    retrieving model instances.
    """
    if select_related_for is None:
        select_related_for = []

    # Group content object pks by their content type pks
    objects = {}
    for item in tagged_items:
        objects.setdefault(item.content_type_id, []).append(item.object_id)

    # Retrieve content types and content objects in bulk
    content_types = ContentType._default_manager.in_bulk(objects.keys())
    for content_type_pk, object_pks in objects.iteritems():
        model = content_types[content_type_pk].model_class()
        if content_types[content_type_pk].model in select_related_for:
            objects[content_type_pk] = model._default_manager.select_related(
                ).in_bulk(object_pks)
        else:
            objects[content_type_pk] = model._default_manager.in_bulk(
                object_pks)

    # Set content types and content objects in the appropriate cache
    # attributes, so accessing the 'content_type' and 'object'
    # attributes on each tagged item won't result in further database
    # hits.
    for item in tagged_items:
        item._object_cache = objects[item.content_type_id][item.object_id]
        item._content_type_cache = content_types[item.content_type_id]



import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tagging', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taggeditem',
            name='content_type',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='contenttypes.ContentType',
                verbose_name='content type'),
        ),
        migrations.AlterField(
            model_name='taggeditem',
            name='tag',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='items',
                to='tagging.Tag',
                verbose_name='tag'),
        ),
    ]



"""Tests settings"""
import os

SECRET_KEY = 'secret-key'

DATABASES = {
    'default': {
        'NAME': 'tagging.db',
        'ENGINE': 'django.db.backends.sqlite3'
    }
}

DATABASE_ENGINE = os.environ.get('DATABASE_ENGINE')
if DATABASE_ENGINE == 'postgres':
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql_psycopg2',
            'NAME': 'tagging',
            'USER': 'postgres',
            'HOST': 'localhost'
        }
    }
elif DATABASE_ENGINE == 'mysql':
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.mysql',
            'NAME': 'zinnia',
            'USER': 'root',
            'HOST': 'localhost',
            'TEST': {
                'COLLATION': 'utf8_general_ci'
            }
        }
    }

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.sessions',
    'django.contrib.contenttypes',
    'tagging',
    'tagging.tests',
]



"""Test urls for tagging."""
from django.conf.urls import url

from tagging.tests.models import Article
from tagging.views import TaggedObjectList


class StaticTaggedObjectList(TaggedObjectList):
    tag = 'static'
    queryset = Article.objects.all()


urlpatterns = [
    url(r'^static/$', StaticTaggedObjectList.as_view()),
    url(r'^static/related/$', StaticTaggedObjectList.as_view(
        related_tags=True)),
    url(r'^no-tag/$', TaggedObjectList.as_view(model=Article)),
    url(r'^no-query-no-model/$', TaggedObjectList.as_view()),
    url(r'^(?P<tag>[^/]+(?u))/$', TaggedObjectList.as_view(model=Article)),
]



"""
Tests utils for tagging.
"""
from django.template import Origin
from django.template.loaders.base import Loader


class VoidLoader(Loader):
    """
    Template loader which is always returning
    an empty template.
    """
    is_usable = True
    _accepts_engine_in_init = True

    def get_template_sources(self, template_name):
        yield Origin(
            name='voidloader',
            template_name=template_name,
            loader=self)

    def get_contents(self, origin):
        return ''

    def load_template_source(self, template_name, template_dirs=None):
        return ('', 'voidloader:%s' % template_name)



"""
Models and managers for tagging.
"""
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import connection
from django.db import models
from django.utils.encoding import smart_text
from django.utils.translation import gettext_lazy as _

from tagging import settings
from tagging.utils import LOGARITHMIC
from tagging.utils import calculate_cloud
from tagging.utils import get_queryset_and_model
from tagging.utils import get_tag_list
from tagging.utils import parse_tag_input


qn = connection.ops.quote_name


############
# Managers #
############

class TagManager(models.Manager):

    def update_tags(self, obj, tag_names):
        """
        Update tags associated with an object.
        """
        ctype = ContentType.objects.get_for_model(obj)
        current_tags = list(self.filter(items__content_type__pk=ctype.pk,
                                        items__object_id=str(obj.pk)))
        updated_tag_names = parse_tag_input(tag_names)
        if settings.FORCE_LOWERCASE_TAGS:
            updated_tag_names = [t.lower() for t in updated_tag_names]

        # Remove tags which no longer apply
        tags_for_removal = [tag for tag in current_tags
                            if tag.name not in updated_tag_names]
        if len(tags_for_removal):
            TaggedItem._default_manager.filter(
                content_type__pk=ctype.pk,
                object_id=str(obj.pk),
                tag__in=tags_for_removal).delete()
        # Add new tags
        current_tag_names = [tag.name for tag in current_tags]
        for tag_name in updated_tag_names:
            if tag_name not in current_tag_names:
                tag, created = self.get_or_create(name=tag_name)
                TaggedItem._default_manager.get_or_create(
                    content_type_id=ctype.pk,
                    object_id=obj.pk,
                    tag=tag,
                )

    def add_tag(self, obj, tag_name):
        """
        Associates the given object with a tag.
        """
        tag_names = parse_tag_input(tag_name)
        if not len(tag_names):
            raise AttributeError(
                _('No tags were given: "%s".') % tag_name)
        if len(tag_names) > 1:
            raise AttributeError(
                _('Multiple tags were given: "%s".') % tag_name)
        tag_name = tag_names[0]
        if settings.FORCE_LOWERCASE_TAGS:
            tag_name = tag_name.lower()
        tag, created = self.get_or_create(name=tag_name)
        ctype = ContentType.objects.get_for_model(obj)
        TaggedItem._default_manager.get_or_create(
            tag=tag, content_type=ctype, object_id=str(obj.pk))

    def get_for_object(self, obj):
        """
        Create a queryset matching all tags associated with the given
        object.
        """
        ctype = ContentType.objects.get_for_model(obj)
        return self.filter(items__content_type__pk=ctype.pk,
                           items__object_id=str(obj.pk))

    def _get_usage(self, model, counts=False, min_count=None,
                   extra_joins=None, extra_criteria=None, params=None):
        """
        Perform the custom SQL query for ``usage_for_model`` and
        ``usage_for_queryset``.
        """
        if min_count is not None:
            counts = True

        model_table = qn(model._meta.db_table)
        model_pk = '%s.%s::text' % (model_table, qn(model._meta.pk.column))
        query = """
        SELECT DISTINCT %(tag)s.id, %(tag)s.name%(count_sql)s
        FROM
            %(tag)s
            INNER JOIN %(tagged_item)s
                ON %(tag)s.id = %(tagged_item)s.tag_id
            INNER JOIN %(model)s
                ON %(tagged_item)s.object_id = %(model_pk)s
            %%s
        WHERE %(tagged_item)s.content_type_id = %(content_type_id)s
            %%s
        GROUP BY %(tag)s.id, %(tag)s.name
        %%s
        ORDER BY %(tag)s.name ASC""" % {
            'tag': qn(self.model._meta.db_table),
            'count_sql': counts and (', COUNT(%s)' % model_pk) or '',
            'tagged_item': qn(TaggedItem._meta.db_table),
            'model': model_table,
            'model_pk': model_pk,
            'content_type_id': ContentType.objects.get_for_model(model).pk,
        }

        min_count_sql = ''
        if min_count is not None:
            min_count_sql = 'HAVING COUNT(%s) >= %%s' % model_pk
            params.append(min_count)

        cursor = connection.cursor()
        cursor.execute(query % (extra_joins, extra_criteria, min_count_sql),
                       params)
        tags = []
        for row in cursor.fetchall():
            t = self.model(*row[:2])
            if counts:
                t.count = row[2]
            tags.append(t)
        return tags

    def usage_for_model(self, model, counts=False, min_count=None,
                        filters=None):
        """
        Obtain a list of tags associated with instances of the given
        Model class.

        If ``counts`` is True, a ``count`` attribute will be added to
        each tag, indicating how many times it has been used against
        the Model class in question.

        If ``min_count`` is given, only tags which have a ``count``
        greater than or equal to ``min_count`` will be returned.
        Passing a value for ``min_count`` implies ``counts=True``.

        To limit the tags (and counts, if specified) returned to those
        used by a subset of the Model's instances, pass a dictionary
        of field lookups to be applied to the given Model as the
        ``filters`` argument.
        """
        if filters is None:
            filters = {}

        queryset = model._default_manager.filter()
        for f in filters.items():
            queryset.query.add_filter(f)
        usage = self.usage_for_queryset(queryset, counts, min_count)

        return usage

    def usage_for_queryset(self, queryset, counts=False, min_count=None):
        """
        Obtain a list of tags associated with instances of a model
        contained in the given queryset.

        If ``counts`` is True, a ``count`` attribute will be added to
        each tag, indicating how many times it has been used against
        the Model class in question.

        If ``min_count`` is given, only tags which have a ``count``
        greater than or equal to ``min_count`` will be returned.
        Passing a value for ``min_count`` implies ``counts=True``.
        """
        compiler = queryset.query.get_compiler(using=queryset.db)
        where, params = compiler.compile(queryset.query.where)
        extra_joins = ' '.join(compiler.get_from_clause()[0][1:])

        if where:
            extra_criteria = 'AND %s' % where
        else:
            extra_criteria = ''
        return self._get_usage(queryset.model, counts, min_count,
                               extra_joins, extra_criteria, params)

    def related_for_model(self, tags, model, counts=False, min_count=None):
        """
        Obtain a list of tags related to a given list of tags - that
        is, other tags used by items which have all the given tags.

        If ``counts`` is True, a ``count`` attribute will be added to
        each tag, indicating the number of items which have it in
        addition to the given list of tags.

        If ``min_count`` is given, only tags which have a ``count``
        greater than or equal to ``min_count`` will be returned.
        Passing a value for ``min_count`` implies ``counts=True``.
        """
        if min_count is not None:
            counts = True

        tags = get_tag_list(tags)
        tag_count = len(tags)
        tagged_item_table = qn(TaggedItem._meta.db_table)
        query = """
        SELECT %(tag)s.id, %(tag)s.name%(count_sql)s
        FROM %(tagged_item)s INNER JOIN %(tag)s ON
             %(tagged_item)s.tag_id = %(tag)s.id
        WHERE %(tagged_item)s.content_type_id = %(content_type_id)s
          AND %(tagged_item)s.object_id IN
          (
              SELECT %(tagged_item)s.object_id
              FROM %(tagged_item)s, %(tag)s
              WHERE %(tagged_item)s.content_type_id = %(content_type_id)s
                AND %(tag)s.id = %(tagged_item)s.tag_id
                AND %(tag)s.id IN (%(tag_id_placeholders)s)
              GROUP BY %(tagged_item)s.object_id
              HAVING COUNT(%(tagged_item)s.object_id) = %(tag_count)s
          )
          AND %(tag)s.id NOT IN (%(tag_id_placeholders)s)
        GROUP BY %(tag)s.id, %(tag)s.name
        %(min_count_sql)s
        ORDER BY %(tag)s.name ASC""" % {
            'tag': qn(self.model._meta.db_table),
            'count_sql': counts and ', COUNT(%s.object_id)' %
                tagged_item_table or '',
            'tagged_item': tagged_item_table,
            'content_type_id': ContentType.objects.get_for_model(model).pk,
            'tag_id_placeholders': ','.join(['%s'] * tag_count),
            'tag_count': tag_count,
            'min_count_sql': min_count is not None and (
                'HAVING COUNT(%s.object_id) >= %%s' % tagged_item_table) or '',
        }

        params = [tag.pk for tag in tags] * 2
        if min_count is not None:
            params.append(min_count)

        cursor = connection.cursor()
        cursor.execute(query, params)
        related = []
        for row in cursor.fetchall():
            tag = self.model(*row[:2])
            if counts is True:
                tag.count = row[2]
            related.append(tag)
        return related

    def cloud_for_model(self, model, steps=4, distribution=LOGARITHMIC,
                        filters=None, min_count=None):
        """
        Obtain a list of tags associated with instances of the given
        Model, giving each tag a ``count`` attribute indicating how
        many times it has been used and a ``font_size`` attribute for
        use in displaying a tag cloud.

        ``steps`` defines the range of font sizes - ``font_size`` will
        be an integer between 1 and ``steps`` (inclusive).

        ``distribution`` defines the type of font size distribution
        algorithm which will be used - logarithmic or linear. It must
        be either ``tagging.utils.LOGARITHMIC`` or
        ``tagging.utils.LINEAR``.

        To limit the tags displayed in the cloud to those associated
        with a subset of the Model's instances, pass a dictionary of
        field lookups to be applied to the given Model as the
        ``filters`` argument.

        To limit the tags displayed in the cloud to those with a
        ``count`` greater than or equal to ``min_count``, pass a value
        for the ``min_count`` argument.
        """
        tags = list(self.usage_for_model(model, counts=True, filters=filters,
                                         min_count=min_count))
        return calculate_cloud(tags, steps, distribution)


class TaggedItemManager(models.Manager):
    """
    FIXME There's currently no way to get the ``GROUP BY`` and ``HAVING``
          SQL clauses required by many of this manager's methods into
          Django's ORM.

          For now, we manually execute a query to retrieve the PKs of
          objects we're interested in, then use the ORM's ``__in``
          lookup to return a ``QuerySet``.

          Now that the queryset-refactor branch is in the trunk, this can be
          tidied up significantly.
    """

    def get_by_model(self, queryset_or_model, tags):
        """
        Create a ``QuerySet`` containing instances of the specified
        model associated with a given tag or list of tags.
        """
        tags = get_tag_list(tags)
        tag_count = len(tags)
        if tag_count == 0:
            # No existing tags were given
            queryset, model = get_queryset_and_model(queryset_or_model)
            return model._default_manager.none()
        elif tag_count == 1:
            # Optimisation for single tag - fall through to the simpler
            # query below.
            tag = tags[0]
        else:
            return self.get_intersection_by_model(queryset_or_model, tags)

        queryset, model = get_queryset_and_model(queryset_or_model)
        content_type = ContentType.objects.get_for_model(model)
        opts = self.model._meta
        tagged_item_table = qn(opts.db_table)
        return queryset.extra(
            tables=[opts.db_table],
            where=[
                '%s.content_type_id = %%s' % tagged_item_table,
                '%s.tag_id = %%s' % tagged_item_table,
                '%s.%s::text = %s.object_id' % (qn(model._meta.db_table),
                                                qn(model._meta.pk.column),
                                                tagged_item_table)
            ],
            params=[content_type.pk, tag.pk],
        )

    def get_intersection_by_model(self, queryset_or_model, tags):
        """
        Create a ``QuerySet`` containing instances of the specified
        model associated with *all* of the given list of tags.
        """
        tags = get_tag_list(tags)
        tag_count = len(tags)
        queryset, model = get_queryset_and_model(queryset_or_model)

        if not tag_count:
            return model._default_manager.none()

        model_table = qn(model._meta.db_table)
        # This query selects the ids of all objects which have all the
        # given tags.
        query = """
        SELECT %(model_pk)s
        FROM %(model)s, %(tagged_item)s
        WHERE %(tagged_item)s.content_type_id = %(content_type_id)s
          AND %(tagged_item)s.tag_id IN (%(tag_id_placeholders)s)
          AND %(model_pk)s = %(tagged_item)s.object_id
        GROUP BY %(model_pk)s
        HAVING COUNT(%(model_pk)s) = %(tag_count)s""" % {
            'model_pk': '%s.%s::text' % (model_table, qn(model._meta.pk.column)),
            'model': model_table,
            'tagged_item': qn(self.model._meta.db_table),
            'content_type_id': ContentType.objects.get_for_model(model).pk,
            'tag_id_placeholders': ','.join(['%s'] * tag_count),
            'tag_count': tag_count,
        }

        cursor = connection.cursor()
        cursor.execute(query, [tag.pk for tag in tags])
        object_ids = [row[0] for row in cursor.fetchall()]
        if len(object_ids) > 0:
            return queryset.filter(pk__in=object_ids)
        else:
            return model._default_manager.none()

    def get_union_by_model(self, queryset_or_model, tags):
        """
        Create a ``QuerySet`` containing instances of the specified
        model associated with *any* of the given list of tags.
        """
        tags = get_tag_list(tags)
        tag_count = len(tags)
        queryset, model = get_queryset_and_model(queryset_or_model)

        if not tag_count:
            return model._default_manager.none()

        model_table = qn(model._meta.db_table)
        # This query selects the ids of all objects which have any of
        # the given tags.
        query = """
        SELECT %(model_pk)s
        FROM %(model)s, %(tagged_item)s
        WHERE %(tagged_item)s.content_type_id = %(content_type_id)s
          AND %(tagged_item)s.tag_id IN (%(tag_id_placeholders)s)
          AND %(model_pk)s = %(tagged_item)s.object_id
        GROUP BY %(model_pk)s""" % {
            'model_pk': '%s.%s::text' % (model_table, qn(model._meta.pk.column)),
            'model': model_table,
            'tagged_item': qn(self.model._meta.db_table),
            'content_type_id': ContentType.objects.get_for_model(model).pk,
            'tag_id_placeholders': ','.join(['%s'] * tag_count),
        }

        cursor = connection.cursor()
        cursor.execute(query, [tag.pk for tag in tags])
        object_ids = [row[0] for row in cursor.fetchall()]
        if len(object_ids) > 0:
            return queryset.filter(pk__in=object_ids)
        else:
            return model._default_manager.none()

    def get_related(self, obj, queryset_or_model, num=None):
        """
        Retrieve a list of instances of the specified model which share
        tags with the model instance ``obj``, ordered by the number of
        shared tags in descending order.

        If ``num`` is given, a maximum of ``num`` instances will be
        returned.
        """
        queryset, model = get_queryset_and_model(queryset_or_model)
        model_table = qn(model._meta.db_table)
        content_type = ContentType.objects.get_for_model(obj)
        related_content_type = ContentType.objects.get_for_model(model)
        query = """
        SELECT %(model_pk)s, COUNT(related_tagged_item.object_id) AS %(count)s
        FROM %(model)s, %(tagged_item)s, %(tag)s,
             %(tagged_item)s related_tagged_item
        WHERE %(tagged_item)s.object_id = %%s
          AND %(tagged_item)s.content_type_id = %(content_type_id)s
          AND %(tag)s.id = %(tagged_item)s.tag_id
          AND related_tagged_item.content_type_id = %(related_content_type_id)s
          AND related_tagged_item.tag_id = %(tagged_item)s.tag_id
          AND %(model_pk)s = related_tagged_item.object_id"""
        if content_type.pk == related_content_type.pk:
            # Exclude the given instance itself if determining related
            # instances for the same model.
            query += """
          AND related_tagged_item.object_id != %(tagged_item)s.object_id"""
        query += """
        GROUP BY %(model_pk)s
        ORDER BY %(count)s DESC
        %(limit_offset)s"""
        tagging_table = qn(self.model._meta.get_field(
            'tag').remote_field.model._meta.db_table)
        query = query % {
            'model_pk': '%s.%s::text' % (model_table, qn(model._meta.pk.column)),
            'count': qn('count'),
            'model': model_table,
            'tagged_item': qn(self.model._meta.db_table),
            'tag': tagging_table,
            'content_type_id': content_type.pk,
            'related_content_type_id': related_content_type.pk,
            # Hardcoding this for now just to get tests working again - this
            # should now be handled by the query object.
            'limit_offset': num is not None and 'LIMIT %s' or '',
        }

        cursor = connection.cursor()
        params = [str(obj.pk)]
        if num is not None:
            params.append(num)
        cursor.execute(query, params)
        object_ids = [row[0] for row in cursor.fetchall()]
        if len(object_ids) > 0:
            # Use in_bulk here instead of an id__in lookup,
            # because id__in would clobber the ordering.
            object_dict = queryset.in_bulk(object_ids)
            return [object_dict[object_id] for object_id in object_ids
                    if object_id in object_dict]
        else:
            return []


##########
# Models #
##########

class Tag(models.Model):
    """
    A tag.
    """
    name = models.CharField(
        _('name'), max_length=settings.MAX_TAG_LENGTH,
        unique=True, db_index=True)

    objects = TagManager()

    class Meta:
        ordering = ('name',)
        verbose_name = _('tag')
        verbose_name_plural = _('tags')

    def __str__(self):
        return self.name


class TaggedItem(models.Model):
    """
    Holds the relationship between a tag and the item being tagged.
    """
    tag = models.ForeignKey(
        Tag,
        verbose_name=_('tag'),
        related_name='items',
        on_delete=models.CASCADE)

    content_type = models.ForeignKey(
        ContentType,
        verbose_name=_('content type'),
        on_delete=models.CASCADE)

    object_id = models.TextField(
        _('object id'),
        db_index=True)

    object = GenericForeignKey(
        'content_type', 'object_id')

    objects = TaggedItemManager()

    created_at = models.DateTimeField(auto_now_add=True, db_index=True, null=True)
    created_user = models.IntegerField(null=True, blank=True)

    class Meta:
        # Enforce unique tag association per object
        unique_together = (('tag', 'content_type', 'object_id'),)
        verbose_name = _('tagged item')
        verbose_name_plural = _('tagged items')

    def __str__(self):
        return '%s [%s]' % (smart_text(self.object), smart_text(self.tag))



from django.db import models

from tagging.fields import TagField


class Perch(models.Model):
    size = models.IntegerField()
    smelly = models.BooleanField(default=True)


class Parrot(models.Model):
    state = models.CharField(max_length=50)
    perch = models.ForeignKey(Perch, null=True,
                              on_delete=models.CASCADE)

    def __str__(self):
        return self.state

    class Meta:
        ordering = ['state']


class Link(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']


class Article(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']


class FormTest(models.Model):
    tags = TagField('Test', help_text='Test')


class FormTestNull(models.Model):
    tags = TagField(null=True)


class FormMultipleFieldTest(models.Model):
    tagging_field = TagField('Test', help_text='Test')
    name = models.CharField(max_length=50)



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tagging', '0004_auto_20170802_1747'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tag',
            name='name',
            field=models.CharField(
                unique=True,
                max_length=getattr(settings, 'MAX_TAG_LENGTH', 50),
                verbose_name='name',
                db_index=True),
        ),
    ]



import os

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q
from django.test import TestCase
from django.test.utils import override_settings

from tagging import settings
from tagging.forms import TagAdminForm
from tagging.forms import TagField
from tagging.models import Tag
from tagging.models import TaggedItem
from tagging.tests.models import Article
from tagging.tests.models import FormMultipleFieldTest
from tagging.tests.models import FormTest
from tagging.tests.models import FormTestNull
from tagging.tests.models import Link
from tagging.tests.models import Parrot
from tagging.tests.models import Perch
from tagging.utils import LINEAR
from tagging.utils import LOGARITHMIC
from tagging.utils import _calculate_tag_weight
from tagging.utils import calculate_cloud
from tagging.utils import edit_string_for_tags
from tagging.utils import get_tag
from tagging.utils import get_tag_list
from tagging.utils import parse_tag_input

#############
# Utilities #
#############


class TestParseTagInput(TestCase):
    def test_with_simple_space_delimited_tags(self):
        """ Test with simple space-delimited tags. """

        self.assertEqual(parse_tag_input('one'), ['one'])
        self.assertEqual(parse_tag_input('one two'), ['one', 'two'])
        self.assertEqual(parse_tag_input('one one two two'), ['one', 'two'])
        self.assertEqual(parse_tag_input('one two three'),
                         ['one', 'three', 'two'])

    def test_with_comma_delimited_multiple_words(self):
        """ Test with comma-delimited multiple words.
            An unquoted comma in the input will trigger this. """

        self.assertEqual(parse_tag_input(',one'), ['one'])
        self.assertEqual(parse_tag_input(',one two'), ['one two'])
        self.assertEqual(parse_tag_input(',one two three'), ['one two three'])
        self.assertEqual(parse_tag_input('a-one, a-two and a-three'),
                         ['a-one', 'a-two and a-three'])

    def test_with_double_quoted_multiple_words(self):
        """ Test with double-quoted multiple words.
            A completed quote will trigger this.  Unclosed quotes are ignored.
        """

        self.assertEqual(parse_tag_input('"one'), ['one'])
        self.assertEqual(parse_tag_input('"one two'), ['one', 'two'])
        self.assertEqual(parse_tag_input('"one two three'),
                         ['one', 'three', 'two'])
        self.assertEqual(parse_tag_input('"one two"'), ['one two'])
        self.assertEqual(parse_tag_input('a-one "a-two and a-three"'),
                         ['a-one', 'a-two and a-three'])

    def test_with_no_loose_commas(self):
        """ Test with no loose commas -- split on spaces. """
        self.assertEqual(parse_tag_input('one two "thr,ee"'),
                         ['one', 'thr,ee', 'two'])

    def test_with_loose_commas(self):
        """ Loose commas - split on commas """
        self.assertEqual(parse_tag_input('"one", two three'),
                         ['one', 'two three'])

    def test_tags_with_double_quotes_can_contain_commas(self):
        """ Double quotes can contain commas """
        self.assertEqual(parse_tag_input('a-one "a-two, and a-three"'),
                         ['a-one', 'a-two, and a-three'])
        self.assertEqual(parse_tag_input('"two", one, one, two, "one"'),
                         ['one', 'two'])
        self.assertEqual(parse_tag_input('two", one'),
                         ['one', 'two'])

    def test_with_naughty_input(self):
        """ Test with naughty input. """
        # Bad users! Naughty users!
        self.assertEqual(parse_tag_input(None), [])
        self.assertEqual(parse_tag_input(''), [])
        self.assertEqual(parse_tag_input('"'), [])
        self.assertEqual(parse_tag_input('""'), [])
        self.assertEqual(parse_tag_input('"' * 7), [])
        self.assertEqual(parse_tag_input(',,,,,,'), [])
        self.assertEqual(parse_tag_input('",",",",",",","'), [','])
        self.assertEqual(parse_tag_input('a-one "a-two" and "a-three'),
                         ['a-one', 'a-three', 'a-two', 'and'])


class TestNormalisedTagListInput(TestCase):
    def setUp(self):
        self.toast = Tag.objects.create(name='toast')
        self.cheese = Tag.objects.create(name='cheese')

    def test_single_tag_object_as_input(self):
        self.assertEqual(get_tag_list(self.cheese), [self.cheese])

    def test_space_delimeted_string_as_input(self):
        ret = get_tag_list('cheese toast')
        self.assertEqual(len(ret), 2)
        self.assertTrue(self.cheese in ret)
        self.assertTrue(self.toast in ret)

    def test_comma_delimeted_string_as_input(self):
        ret = get_tag_list('cheese,toast')
        self.assertEqual(len(ret), 2)
        self.assertTrue(self.cheese in ret)
        self.assertTrue(self.toast in ret)

    def test_with_empty_list(self):
        self.assertEqual(get_tag_list([]), [])

    def test_list_of_two_strings(self):
        ret = get_tag_list(['cheese', 'toast'])
        self.assertEqual(len(ret), 2)
        self.assertTrue(self.cheese in ret)
        self.assertTrue(self.toast in ret)

    def test_list_of_tag_primary_keys(self):
        ret = get_tag_list([self.cheese.id, self.toast.id])
        self.assertEqual(len(ret), 2)
        self.assertTrue(self.cheese in ret)
        self.assertTrue(self.toast in ret)

    def test_list_of_strings_with_strange_nontag_string(self):
        ret = get_tag_list(['cheese', 'toast', 'ŠĐĆŽćžšđ'])
        self.assertEqual(len(ret), 2)
        self.assertTrue(self.cheese in ret)
        self.assertTrue(self.toast in ret)

    def test_list_of_tag_instances(self):
        ret = get_tag_list([self.cheese, self.toast])
        self.assertEqual(len(ret), 2)
        self.assertTrue(self.cheese in ret)
        self.assertTrue(self.toast in ret)

    def test_tuple_of_instances(self):
        ret = get_tag_list((self.cheese, self.toast))
        self.assertEqual(len(ret), 2)
        self.assertTrue(self.cheese in ret)
        self.assertTrue(self.toast in ret)

    def test_with_tag_filter(self):
        ret = get_tag_list(Tag.objects.filter(name__in=['cheese', 'toast']))
        self.assertEqual(len(ret), 2)
        self.assertTrue(self.cheese in ret)
        self.assertTrue(self.toast in ret)

    def test_with_invalid_input_mix_of_string_and_instance(self):
        try:
            get_tag_list(['cheese', self.toast])
        except ValueError as ve:
            self.assertEqual(
                str(ve),
                'If a list or tuple of tags is provided, they must all '
                'be tag names, Tag objects or Tag ids.')
        except Exception as e:
            raise self.failureException(
                'the wrong type of exception was raised: type [%s] value [%]' %
                (str(type(e)), str(e)))
        else:
            raise self.failureException(
                'a ValueError exception was supposed to be raised!')

    def test_with_invalid_input(self):
        try:
            get_tag_list(29)
        except ValueError as ve:
            self.assertEqual(str(ve), 'The tag input given was invalid.')
        except Exception as e:
            print('--', e)
            raise self.failureException(
                'the wrong type of exception was raised: '
                'type [%s] value [%s]' % (str(type(e)), str(e)))
        else:
            raise self.failureException(
                'a ValueError exception was supposed to be raised!')

    def test_with_tag_instance(self):
        self.assertEqual(get_tag(self.cheese), self.cheese)

    def test_with_string(self):
        self.assertEqual(get_tag('cheese'), self.cheese)

    def test_with_primary_key(self):
        self.assertEqual(get_tag(self.cheese.id), self.cheese)

    def test_nonexistent_tag(self):
        self.assertEqual(get_tag('mouse'), None)


class TestCalculateCloud(TestCase):
    def setUp(self):
        self.tags = []
        for line in open(os.path.join(os.path.dirname(__file__),
                                      'tags.txt')).readlines():
            name, count = line.rstrip().split()
            tag = Tag(name=name)
            tag.count = int(count)
            self.tags.append(tag)

    def test_default_distribution(self):
        sizes = {}
        for tag in calculate_cloud(self.tags, steps=5):
            sizes[tag.font_size] = sizes.get(tag.font_size, 0) + 1

        # This isn't a pre-calculated test, just making sure it's consistent
        self.assertEqual(sizes[1], 48)
        self.assertEqual(sizes[2], 30)
        self.assertEqual(sizes[3], 19)
        self.assertEqual(sizes[4], 15)
        self.assertEqual(sizes[5], 10)

    def test_linear_distribution(self):
        sizes = {}
        for tag in calculate_cloud(self.tags, steps=5, distribution=LINEAR):
            sizes[tag.font_size] = sizes.get(tag.font_size, 0) + 1

        # This isn't a pre-calculated test, just making sure it's consistent
        self.assertEqual(sizes[1], 97)
        self.assertEqual(sizes[2], 12)
        self.assertEqual(sizes[3], 7)
        self.assertEqual(sizes[4], 2)
        self.assertEqual(sizes[5], 4)

    def test_invalid_distribution(self):
        try:
            calculate_cloud(self.tags, steps=5, distribution='cheese')
        except ValueError as ve:
            self.assertEqual(
                str(ve), 'Invalid distribution algorithm specified: cheese.')
        except Exception as e:
            raise self.failureException(
                'the wrong type of exception was raised: '
                'type [%s] value [%s]' % (str(type(e)), str(e)))
        else:
            raise self.failureException(
                'a ValueError exception was supposed to be raised!')

    def test_calculate_tag_weight(self):
        self.assertEqual(
            _calculate_tag_weight(10, 20, LINEAR),
            10)
        self.assertEqual(
            _calculate_tag_weight(10, 20, LOGARITHMIC),
            15.37243573680482)

    def test_calculate_tag_weight_invalid_size(self):
        self.assertEqual(
            _calculate_tag_weight(10, 10, LOGARITHMIC),
            10.0)
        self.assertEqual(
            _calculate_tag_weight(26, 26, LOGARITHMIC),
            26.0)

###########
# Tagging #
###########


class TestBasicTagging(TestCase):
    def setUp(self):
        self.dead_parrot = Parrot.objects.create(state='dead')

    def test_update_tags(self):
        Tag.objects.update_tags(self.dead_parrot, 'foo,bar,"ter"')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(get_tag('foo') in tags)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('ter') in tags)

        Tag.objects.update_tags(self.dead_parrot, '"foo" bar "baz"')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('baz') in tags)
        self.assertTrue(get_tag('foo') in tags)

    def test_add_tag(self):
        # start off in a known, mildly interesting state
        Tag.objects.update_tags(self.dead_parrot, 'foo bar baz')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('baz') in tags)
        self.assertTrue(get_tag('foo') in tags)

        # try to add a tag that already exists
        Tag.objects.add_tag(self.dead_parrot, 'foo')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('baz') in tags)
        self.assertTrue(get_tag('foo') in tags)

        # now add a tag that doesn't already exist
        Tag.objects.add_tag(self.dead_parrot, 'zip')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 4)
        self.assertTrue(get_tag('zip') in tags)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('baz') in tags)
        self.assertTrue(get_tag('foo') in tags)

    def test_add_tag_invalid_input_no_tags_specified(self):
        # start off in a known, mildly interesting state
        Tag.objects.update_tags(self.dead_parrot, 'foo bar baz')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('baz') in tags)
        self.assertTrue(get_tag('foo') in tags)

        try:
            Tag.objects.add_tag(self.dead_parrot, '    ')
        except AttributeError as ae:
            self.assertEqual(str(ae), 'No tags were given: "    ".')
        except Exception as e:
            raise self.failureException(
                'the wrong type of exception was raised: '
                'type [%s] value [%s]' % (str(type(e)), str(e)))
        else:
            raise self.failureException(
                'an AttributeError exception was supposed to be raised!')

    def test_add_tag_invalid_input_multiple_tags_specified(self):
        # start off in a known, mildly interesting state
        Tag.objects.update_tags(self.dead_parrot, 'foo bar baz')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('baz') in tags)
        self.assertTrue(get_tag('foo') in tags)

        try:
            Tag.objects.add_tag(self.dead_parrot, 'one two')
        except AttributeError as ae:
            self.assertEqual(str(ae), 'Multiple tags were given: "one two".')
        except Exception as e:
            raise self.failureException(
                'the wrong type of exception was raised: '
                'type [%s] value [%s]' % (str(type(e)), str(e)))
        else:
            raise self.failureException(
                'an AttributeError exception was supposed to be raised!')

    def test_update_tags_exotic_characters(self):
        # start off in a known, mildly interesting state
        Tag.objects.update_tags(self.dead_parrot, 'foo bar baz')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('baz') in tags)
        self.assertTrue(get_tag('foo') in tags)

        Tag.objects.update_tags(self.dead_parrot, 'ŠĐĆŽćžšđ')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0].name, 'ŠĐĆŽćžšđ')

        Tag.objects.update_tags(self.dead_parrot, '你好')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0].name, '你好')

    def test_unicode_tagged_object(self):
        self.dead_parrot.state = "dëad"
        self.dead_parrot.save()
        Tag.objects.update_tags(self.dead_parrot, 'föo')
        items = TaggedItem.objects.all()
        self.assertEqual(len(items), 1)
        self.assertEqual(str(items[0]), "dëad [föo]")

    def test_update_tags_with_none(self):
        # start off in a known, mildly interesting state
        Tag.objects.update_tags(self.dead_parrot, 'foo bar baz')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(get_tag('bar') in tags)
        self.assertTrue(get_tag('baz') in tags)
        self.assertTrue(get_tag('foo') in tags)

        Tag.objects.update_tags(self.dead_parrot, None)
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 0)


class TestModelTagField(TestCase):
    """ Test the 'tags' field on models. """

    def test_create_with_tags_specified(self):
        f1 = FormTest.objects.create(tags='test3 test2 test1')
        tags = Tag.objects.get_for_object(f1)
        test1_tag = get_tag('test1')
        test2_tag = get_tag('test2')
        test3_tag = get_tag('test3')
        self.assertTrue(None not in (test1_tag, test2_tag, test3_tag))
        self.assertEqual(len(tags), 3)
        self.assertTrue(test1_tag in tags)
        self.assertTrue(test2_tag in tags)
        self.assertTrue(test3_tag in tags)

    def test_update_via_tags_field(self):
        f1 = FormTest.objects.create(tags='test3 test2 test1')
        tags = Tag.objects.get_for_object(f1)
        test1_tag = get_tag('test1')
        test2_tag = get_tag('test2')
        test3_tag = get_tag('test3')
        self.assertTrue(None not in (test1_tag, test2_tag, test3_tag))
        self.assertEqual(len(tags), 3)
        self.assertTrue(test1_tag in tags)
        self.assertTrue(test2_tag in tags)
        self.assertTrue(test3_tag in tags)

        f1.tags = 'test4'
        f1.save()
        tags = Tag.objects.get_for_object(f1)
        test4_tag = get_tag('test4')
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0], test4_tag)

        f1.tags = ''
        f1.save()
        tags = Tag.objects.get_for_object(f1)
        self.assertEqual(len(tags), 0)

    def disabledtest_update_via_tags(self):
        # TODO: make this test working by reverting
        # https://github.com/Fantomas42/django-tagging/commit/bbc7f25ea471dd903f39e08684d84ce59081bdef
        f1 = FormTest.objects.create(tags='one two three')
        Tag.objects.get(name='three').delete()
        t2 = Tag.objects.get(name='two')
        t2.name = 'new'
        t2.save()
        f1again = FormTest.objects.get(pk=f1.pk)
        self.assertFalse('three' in f1again.tags)
        self.assertFalse('two' in f1again.tags)
        self.assertTrue('new' in f1again.tags)

    def test_creation_without_specifying_tags(self):
        f1 = FormTest()
        self.assertEqual(f1.tags, '')

    def test_creation_with_nullable_tags_field(self):
        f1 = FormTestNull()
        self.assertEqual(f1.tags, '')

    def test_fix_update_tag_field_deferred(self):
        """
        Bug introduced in Django 1.10
        the TagField is considered "deferred" on Django 1.10
        because instance.__dict__ is not populated by the TagField
        instance, so it's excluded when updating a model instance.

        Note: this does not append if you only have one TagField
        in your model...
        """
        f1 = FormMultipleFieldTest.objects.create(tagging_field='one two')
        self.assertEqual(f1.tagging_field, 'one two')
        tags = Tag.objects.get_for_object(f1)
        self.assertEqual(len(tags), 2)
        test1_tag = get_tag('one')
        test2_tag = get_tag('two')
        self.assertTrue(test1_tag in tags)
        self.assertTrue(test2_tag in tags)

        f1.tagging_field = f1.tagging_field + ' three'
        f1.save()
        self.assertEqual(f1.tagging_field, 'one two three')
        tags = Tag.objects.get_for_object(f1)
        self.assertEqual(len(tags), 3)
        test3_tag = get_tag('three')
        self.assertTrue(test3_tag in tags)

        f1again = FormMultipleFieldTest.objects.get(pk=f1.pk)
        self.assertEqual(f1again.tagging_field, 'one two three')

        tags = Tag.objects.get_for_object(f1again)
        self.assertEqual(len(tags), 3)


class TestSettings(TestCase):
    def setUp(self):
        self.original_force_lower_case_tags = settings.FORCE_LOWERCASE_TAGS
        self.dead_parrot = Parrot.objects.create(state='dead')

    def tearDown(self):
        settings.FORCE_LOWERCASE_TAGS = self.original_force_lower_case_tags

    def test_force_lowercase_tags(self):
        """ Test forcing tags to lowercase. """

        settings.FORCE_LOWERCASE_TAGS = True

        Tag.objects.update_tags(self.dead_parrot, 'foO bAr Ter')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        foo_tag = get_tag('foo')
        bar_tag = get_tag('bar')
        ter_tag = get_tag('ter')
        self.assertTrue(foo_tag in tags)
        self.assertTrue(bar_tag in tags)
        self.assertTrue(ter_tag in tags)

        Tag.objects.update_tags(self.dead_parrot, 'foO bAr baZ')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        baz_tag = get_tag('baz')
        self.assertEqual(len(tags), 3)
        self.assertTrue(bar_tag in tags)
        self.assertTrue(baz_tag in tags)
        self.assertTrue(foo_tag in tags)

        Tag.objects.add_tag(self.dead_parrot, 'FOO')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 3)
        self.assertTrue(bar_tag in tags)
        self.assertTrue(baz_tag in tags)
        self.assertTrue(foo_tag in tags)

        Tag.objects.add_tag(self.dead_parrot, 'Zip')
        tags = Tag.objects.get_for_object(self.dead_parrot)
        self.assertEqual(len(tags), 4)
        zip_tag = get_tag('zip')
        self.assertTrue(bar_tag in tags)
        self.assertTrue(baz_tag in tags)
        self.assertTrue(foo_tag in tags)
        self.assertTrue(zip_tag in tags)

        f1 = FormTest.objects.create()
        f1.tags = 'TEST5'
        f1.save()
        tags = Tag.objects.get_for_object(f1)
        test5_tag = get_tag('test5')
        self.assertEqual(len(tags), 1)
        self.assertTrue(test5_tag in tags)
        self.assertEqual(f1.tags, 'test5')


class TestTagUsageForModelBaseCase(TestCase):
    def test_tag_usage_for_model_empty(self):
        self.assertEqual(Tag.objects.usage_for_model(Parrot), [])


class TestTagUsageForModel(TestCase):
    def setUp(self):
        parrot_details = (
            ('pining for the fjords', 9, True,  'foo bar'),
            ('passed on',             6, False, 'bar baz ter'),
            ('no more',               4, True,  'foo ter'),
            ('late',                  2, False, 'bar ter'),
        )

        for state, perch_size, perch_smelly, tags in parrot_details:
            perch = Perch.objects.create(size=perch_size, smelly=perch_smelly)
            parrot = Parrot.objects.create(state=state, perch=perch)
            Tag.objects.update_tags(parrot, tags)

    def test_tag_usage_for_model(self):
        tag_usage = Tag.objects.usage_for_model(Parrot, counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 4)
        self.assertTrue(('bar', 3) in relevant_attribute_list)
        self.assertTrue(('baz', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 2) in relevant_attribute_list)
        self.assertTrue(('ter', 3) in relevant_attribute_list)

    def test_tag_usage_for_model_with_min_count(self):
        tag_usage = Tag.objects.usage_for_model(Parrot, min_count=2)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('bar', 3) in relevant_attribute_list)
        self.assertTrue(('foo', 2) in relevant_attribute_list)
        self.assertTrue(('ter', 3) in relevant_attribute_list)

    def test_tag_usage_with_filter_on_model_objects(self):
        tag_usage = Tag.objects.usage_for_model(
            Parrot, counts=True, filters=dict(state='no more'))
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 2)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_model(
            Parrot, counts=True, filters=dict(state__startswith='p'))
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 4)
        self.assertTrue(('bar', 2) in relevant_attribute_list)
        self.assertTrue(('baz', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_model(
            Parrot, counts=True, filters=dict(perch__size__gt=4))
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 4)
        self.assertTrue(('bar', 2) in relevant_attribute_list)
        self.assertTrue(('baz', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_model(
            Parrot, counts=True, filters=dict(perch__smelly=True))
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('bar', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 2) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_model(
            Parrot, min_count=2, filters=dict(perch__smelly=True))
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 1)
        self.assertTrue(('foo', 2) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_model(
            Parrot, filters=dict(perch__size__gt=4))
        relevant_attribute_list = [(tag.name, hasattr(tag, 'counts'))
                                   for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 4)
        self.assertTrue(('bar', False) in relevant_attribute_list)
        self.assertTrue(('baz', False) in relevant_attribute_list)
        self.assertTrue(('foo', False) in relevant_attribute_list)
        self.assertTrue(('ter', False) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_model(
            Parrot, filters=dict(perch__size__gt=99))
        relevant_attribute_list = [(tag.name, hasattr(tag, 'counts'))
                                   for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 0)


class TestTagsRelatedForModel(TestCase):
    def setUp(self):
        parrot_details = (
            ('pining for the fjords', 9, True,  'foo bar'),
            ('passed on',             6, False, 'bar baz ter'),
            ('no more',               4, True,  'foo ter'),
            ('late',                  2, False, 'bar ter'),
        )

        for state, perch_size, perch_smelly, tags in parrot_details:
            perch = Perch.objects.create(size=perch_size, smelly=perch_smelly)
            parrot = Parrot.objects.create(state=state, perch=perch)
            Tag.objects.update_tags(parrot, tags)

    def test_related_for_model_with_tag_query_sets_as_input(self):
        related_tags = Tag.objects.related_for_model(
            Tag.objects.filter(name__in=['bar']), Parrot, counts=True)
        relevant_attribute_list = [(tag.name, tag.count)
                                   for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('baz', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 2) in relevant_attribute_list)

        related_tags = Tag.objects.related_for_model(
            Tag.objects.filter(name__in=['bar']), Parrot, min_count=2)
        relevant_attribute_list = [(tag.name, tag.count)
                                   for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 1)
        self.assertTrue(('ter', 2) in relevant_attribute_list)

        related_tags = Tag.objects.related_for_model(
            Tag.objects.filter(name__in=['bar']), Parrot, counts=False)
        relevant_attribute_list = [tag.name for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue('baz' in relevant_attribute_list)
        self.assertTrue('foo' in relevant_attribute_list)
        self.assertTrue('ter' in relevant_attribute_list)

        related_tags = Tag.objects.related_for_model(
            Tag.objects.filter(name__in=['bar', 'ter']), Parrot, counts=True)
        relevant_attribute_list = [(tag.name, tag.count)
                                   for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 1)
        self.assertTrue(('baz', 1) in relevant_attribute_list)

        related_tags = Tag.objects.related_for_model(
            Tag.objects.filter(name__in=['bar', 'ter', 'baz']),
            Parrot, counts=True)
        relevant_attribute_list = [(tag.name, tag.count)
                                   for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 0)

    def test_related_for_model_with_tag_strings_as_input(self):
        # Once again, with feeling (strings)
        related_tags = Tag.objects.related_for_model(
            'bar', Parrot, counts=True)
        relevant_attribute_list = [(tag.name, tag.count)
                                   for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('baz', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 2) in relevant_attribute_list)

        related_tags = Tag.objects.related_for_model(
            'bar', Parrot, min_count=2)
        relevant_attribute_list = [(tag.name, tag.count)
                                   for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 1)
        self.assertTrue(('ter', 2) in relevant_attribute_list)

        related_tags = Tag.objects.related_for_model(
            'bar', Parrot, counts=False)
        relevant_attribute_list = [tag.name for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue('baz' in relevant_attribute_list)
        self.assertTrue('foo' in relevant_attribute_list)
        self.assertTrue('ter' in relevant_attribute_list)

        related_tags = Tag.objects.related_for_model(
            ['bar', 'ter'], Parrot, counts=True)
        relevant_attribute_list = [(tag.name, tag.count)
                                   for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 1)
        self.assertTrue(('baz', 1) in relevant_attribute_list)

        related_tags = Tag.objects.related_for_model(
            ['bar', 'ter', 'baz'], Parrot, counts=True)
        relevant_attribute_list = [(tag.name, tag.count)
                                   for tag in related_tags]
        self.assertEqual(len(relevant_attribute_list), 0)


class TestTagCloudForModel(TestCase):
    def setUp(self):
        parrot_details = (
            ('pining for the fjords', 9, True,  'foo bar'),
            ('passed on',             6, False, 'bar baz ter'),
            ('no more',               4, True,  'foo ter'),
            ('late',                  2, False, 'bar ter'),
        )

        for state, perch_size, perch_smelly, tags in parrot_details:
            perch = Perch.objects.create(size=perch_size, smelly=perch_smelly)
            parrot = Parrot.objects.create(state=state, perch=perch)
            Tag.objects.update_tags(parrot, tags)

    def test_tag_cloud_for_model(self):
        tag_cloud = Tag.objects.cloud_for_model(Parrot)
        relevant_attribute_list = [(tag.name, tag.count, tag.font_size)
                                   for tag in tag_cloud]
        self.assertEqual(len(relevant_attribute_list), 4)
        self.assertTrue(('bar', 3, 4) in relevant_attribute_list)
        self.assertTrue(('baz', 1, 1) in relevant_attribute_list)
        self.assertTrue(('foo', 2, 2) in relevant_attribute_list)
        self.assertTrue(('ter', 3, 4) in relevant_attribute_list)

    def test_tag_cloud_for_model_filters(self):
        tag_cloud = Tag.objects.cloud_for_model(Parrot,
                                                filters={'state': 'no more'})
        relevant_attribute_list = [(tag.name, tag.count, tag.font_size)
                                   for tag in tag_cloud]
        self.assertEqual(len(relevant_attribute_list), 2)
        self.assertTrue(('foo', 1, 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1, 1) in relevant_attribute_list)

    def test_tag_cloud_for_model_min_count(self):
        tag_cloud = Tag.objects.cloud_for_model(Parrot, min_count=2)
        relevant_attribute_list = [(tag.name, tag.count, tag.font_size)
                                   for tag in tag_cloud]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('bar', 3, 4) in relevant_attribute_list)
        self.assertTrue(('foo', 2, 1) in relevant_attribute_list)
        self.assertTrue(('ter', 3, 4) in relevant_attribute_list)


class TestGetTaggedObjectsByModel(TestCase):
    def setUp(self):
        parrot_details = (
            ('pining for the fjords', 9, True,  'foo bar'),
            ('passed on',             6, False, 'bar baz ter'),
            ('no more',               4, True,  'foo ter'),
            ('late',                  2, False, 'bar ter'),
        )

        for state, perch_size, perch_smelly, tags in parrot_details:
            perch = Perch.objects.create(size=perch_size, smelly=perch_smelly)
            parrot = Parrot.objects.create(state=state, perch=perch)
            Tag.objects.update_tags(parrot, tags)

        self.foo = Tag.objects.get(name='foo')
        self.bar = Tag.objects.get(name='bar')
        self.baz = Tag.objects.get(name='baz')
        self.ter = Tag.objects.get(name='ter')

        self.pining_for_the_fjords_parrot = Parrot.objects.get(
            state='pining for the fjords')
        self.passed_on_parrot = Parrot.objects.get(state='passed on')
        self.no_more_parrot = Parrot.objects.get(state='no more')
        self.late_parrot = Parrot.objects.get(state='late')

    def test_get_by_model_simple(self):
        parrots = TaggedItem.objects.get_by_model(Parrot, self.foo)
        self.assertEqual(len(parrots), 2)
        self.assertTrue(self.no_more_parrot in parrots)
        self.assertTrue(self.pining_for_the_fjords_parrot in parrots)

        parrots = TaggedItem.objects.get_by_model(Parrot, self.bar)
        self.assertEqual(len(parrots), 3)
        self.assertTrue(self.late_parrot in parrots)
        self.assertTrue(self.passed_on_parrot in parrots)
        self.assertTrue(self.pining_for_the_fjords_parrot in parrots)

    def test_get_by_model_intersection(self):
        parrots = TaggedItem.objects.get_by_model(Parrot, [self.foo, self.baz])
        self.assertEqual(len(parrots), 0)

        parrots = TaggedItem.objects.get_by_model(Parrot, [self.foo, self.bar])
        self.assertEqual(len(parrots), 1)
        self.assertTrue(self.pining_for_the_fjords_parrot in parrots)

        parrots = TaggedItem.objects.get_by_model(Parrot, [self.bar, self.ter])
        self.assertEqual(len(parrots), 2)
        self.assertTrue(self.late_parrot in parrots)
        self.assertTrue(self.passed_on_parrot in parrots)

        # Issue 114 - Intersection with non-existant tags
        parrots = TaggedItem.objects.get_intersection_by_model(Parrot, [])
        self.assertEqual(len(parrots), 0)

    def test_get_by_model_with_tag_querysets_as_input(self):
        parrots = TaggedItem.objects.get_by_model(
            Parrot, Tag.objects.filter(name__in=['foo', 'baz']))
        self.assertEqual(len(parrots), 0)

        parrots = TaggedItem.objects.get_by_model(
            Parrot, Tag.objects.filter(name__in=['foo', 'bar']))
        self.assertEqual(len(parrots), 1)
        self.assertTrue(self.pining_for_the_fjords_parrot in parrots)

        parrots = TaggedItem.objects.get_by_model(
            Parrot, Tag.objects.filter(name__in=['bar', 'ter']))
        self.assertEqual(len(parrots), 2)
        self.assertTrue(self.late_parrot in parrots)
        self.assertTrue(self.passed_on_parrot in parrots)

    def test_get_by_model_with_strings_as_input(self):
        parrots = TaggedItem.objects.get_by_model(Parrot, 'foo baz')
        self.assertEqual(len(parrots), 0)

        parrots = TaggedItem.objects.get_by_model(Parrot, 'foo bar')
        self.assertEqual(len(parrots), 1)
        self.assertTrue(self.pining_for_the_fjords_parrot in parrots)

        parrots = TaggedItem.objects.get_by_model(Parrot, 'bar ter')
        self.assertEqual(len(parrots), 2)
        self.assertTrue(self.late_parrot in parrots)
        self.assertTrue(self.passed_on_parrot in parrots)

    def test_get_by_model_with_lists_of_strings_as_input(self):
        parrots = TaggedItem.objects.get_by_model(Parrot, ['foo', 'baz'])
        self.assertEqual(len(parrots), 0)

        parrots = TaggedItem.objects.get_by_model(Parrot, ['foo', 'bar'])
        self.assertEqual(len(parrots), 1)
        self.assertTrue(self.pining_for_the_fjords_parrot in parrots)

        parrots = TaggedItem.objects.get_by_model(Parrot, ['bar', 'ter'])
        self.assertEqual(len(parrots), 2)
        self.assertTrue(self.late_parrot in parrots)
        self.assertTrue(self.passed_on_parrot in parrots)

    def test_get_by_nonexistent_tag(self):
        # Issue 50 - Get by non-existent tag
        parrots = TaggedItem.objects.get_by_model(Parrot, 'argatrons')
        self.assertEqual(len(parrots), 0)

    def test_get_union_by_model(self):
        parrots = TaggedItem.objects.get_union_by_model(Parrot, ['foo', 'ter'])
        self.assertEqual(len(parrots), 4)
        self.assertTrue(self.late_parrot in parrots)
        self.assertTrue(self.no_more_parrot in parrots)
        self.assertTrue(self.passed_on_parrot in parrots)
        self.assertTrue(self.pining_for_the_fjords_parrot in parrots)

        parrots = TaggedItem.objects.get_union_by_model(Parrot, ['bar', 'baz'])
        self.assertEqual(len(parrots), 3)
        self.assertTrue(self.late_parrot in parrots)
        self.assertTrue(self.passed_on_parrot in parrots)
        self.assertTrue(self.pining_for_the_fjords_parrot in parrots)

        # Issue 114 - Union with non-existant tags
        parrots = TaggedItem.objects.get_union_by_model(Parrot, [])
        self.assertEqual(len(parrots), 0)
        parrots = TaggedItem.objects.get_union_by_model(Parrot, ['albert'])
        self.assertEqual(len(parrots), 0)

        Tag.objects.create(name='titi')
        parrots = TaggedItem.objects.get_union_by_model(Parrot, ['titi'])
        self.assertEqual(len(parrots), 0)


class TestGetRelatedTaggedItems(TestCase):
    def setUp(self):
        parrot_details = (
            ('pining for the fjords', 9, True,  'foo bar'),
            ('passed on',             6, False, 'bar baz ter'),
            ('no more',               4, True,  'foo ter'),
            ('late',                  2, False, 'bar ter'),
        )

        for state, perch_size, perch_smelly, tags in parrot_details:
            perch = Perch.objects.create(size=perch_size, smelly=perch_smelly)
            parrot = Parrot.objects.create(state=state, perch=perch)
            Tag.objects.update_tags(parrot, tags)

        self.l1 = Link.objects.create(name='link 1')
        Tag.objects.update_tags(self.l1, 'tag1 tag2 tag3 tag4 tag5')
        self.l2 = Link.objects.create(name='link 2')
        Tag.objects.update_tags(self.l2, 'tag1 tag2 tag3')
        self.l3 = Link.objects.create(name='link 3')
        Tag.objects.update_tags(self.l3, 'tag1')
        self.l4 = Link.objects.create(name='link 4')

        self.a1 = Article.objects.create(name='article 1')
        Tag.objects.update_tags(self.a1, 'tag1 tag2 tag3 tag4')

    def test_get_related_objects_of_same_model(self):
        related_objects = TaggedItem.objects.get_related(self.l1, Link)
        self.assertEqual(len(related_objects), 2)
        self.assertTrue(self.l2 in related_objects)
        self.assertTrue(self.l3 in related_objects)

        related_objects = TaggedItem.objects.get_related(self.l4, Link)
        self.assertEqual(len(related_objects), 0)

    def test_get_related_objects_of_same_model_limited_number_of_results(self):
        # This fails on Oracle because it has no support for a 'LIMIT' clause.
        # See http://bit.ly/1AYNEsa

        # ask for no more than 1 result
        related_objects = TaggedItem.objects.get_related(self.l1, Link, num=1)
        self.assertEqual(len(related_objects), 1)
        self.assertTrue(self.l2 in related_objects)

    def test_get_related_objects_of_same_model_limit_related_items(self):
        related_objects = TaggedItem.objects.get_related(
            self.l1, Link.objects.exclude(name='link 3'))
        self.assertEqual(len(related_objects), 1)
        self.assertTrue(self.l2 in related_objects)

    def test_get_related_objects_of_different_model(self):
        related_objects = TaggedItem.objects.get_related(self.a1, Link)
        self.assertEqual(len(related_objects), 3)
        self.assertTrue(self.l1 in related_objects)
        self.assertTrue(self.l2 in related_objects)
        self.assertTrue(self.l3 in related_objects)

        Tag.objects.update_tags(self.a1, 'tag6')
        related_objects = TaggedItem.objects.get_related(self.a1, Link)
        self.assertEqual(len(related_objects), 0)


class TestTagUsageForQuerySet(TestCase):
    def setUp(self):
        parrot_details = (
            ('pining for the fjords', 9, True,  'foo bar'),
            ('passed on',             6, False, 'bar baz ter'),
            ('no more',               4, True,  'foo ter'),
            ('late',                  2, False, 'bar ter'),
        )

        for state, perch_size, perch_smelly, tags in parrot_details:
            perch = Perch.objects.create(size=perch_size, smelly=perch_smelly)
            parrot = Parrot.objects.create(state=state, perch=perch)
            Tag.objects.update_tags(parrot, tags)

    def test_tag_usage_for_queryset(self):
        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(state='no more'), counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 2)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(state__startswith='p'), counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 4)
        self.assertTrue(('bar', 2) in relevant_attribute_list)
        self.assertTrue(('baz', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(perch__size__gt=4), counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 4)
        self.assertTrue(('bar', 2) in relevant_attribute_list)
        self.assertTrue(('baz', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(perch__smelly=True), counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('bar', 1) in relevant_attribute_list)
        self.assertTrue(('foo', 2) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(perch__smelly=True), min_count=2)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 1)
        self.assertTrue(('foo', 2) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(perch__size__gt=4))
        relevant_attribute_list = [(tag.name, hasattr(tag, 'counts'))
                                   for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 4)
        self.assertTrue(('bar', False) in relevant_attribute_list)
        self.assertTrue(('baz', False) in relevant_attribute_list)
        self.assertTrue(('foo', False) in relevant_attribute_list)
        self.assertTrue(('ter', False) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(perch__size__gt=99))
        relevant_attribute_list = [(tag.name, hasattr(tag, 'counts'))
                                   for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 0)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(Q(perch__size__gt=6) |
                                  Q(state__startswith='l')), counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('bar', 2) in relevant_attribute_list)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(Q(perch__size__gt=6) |
                                  Q(state__startswith='l')), min_count=2)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 1)
        self.assertTrue(('bar', 2) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.filter(Q(perch__size__gt=6) |
                                  Q(state__startswith='l')))
        relevant_attribute_list = [(tag.name, hasattr(tag, 'counts'))
                                   for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('bar', False) in relevant_attribute_list)
        self.assertTrue(('foo', False) in relevant_attribute_list)
        self.assertTrue(('ter', False) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.exclude(state='passed on'), counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 3)
        self.assertTrue(('bar', 2) in relevant_attribute_list)
        self.assertTrue(('foo', 2) in relevant_attribute_list)
        self.assertTrue(('ter', 2) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.exclude(state__startswith='p'), min_count=2)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 1)
        self.assertTrue(('ter', 2) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.exclude(Q(perch__size__gt=6) |
                                   Q(perch__smelly=False)), counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 2)
        self.assertTrue(('foo', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)

        tag_usage = Tag.objects.usage_for_queryset(
            Parrot.objects.exclude(perch__smelly=True).filter(
                state__startswith='l'), counts=True)
        relevant_attribute_list = [(tag.name, tag.count) for tag in tag_usage]
        self.assertEqual(len(relevant_attribute_list), 2)
        self.assertTrue(('bar', 1) in relevant_attribute_list)
        self.assertTrue(('ter', 1) in relevant_attribute_list)


################
# Model Fields #
################

class TestTagFieldInForms(TestCase):
    def test_tag_field_in_modelform(self):
        # Ensure that automatically created forms use TagField
        class TestForm(forms.ModelForm):
            class Meta:
                model = FormTest
                fields = forms.ALL_FIELDS

        form = TestForm()
        self.assertEqual(form.fields['tags'].__class__.__name__, 'TagField')

    def test_recreation_of_tag_list_string_representations(self):
        plain = Tag.objects.create(name='plain')
        spaces = Tag.objects.create(name='spa ces')
        comma = Tag.objects.create(name='com,ma')
        self.assertEqual(edit_string_for_tags([plain]), 'plain')
        self.assertEqual(edit_string_for_tags([spaces]), '"spa ces"')
        self.assertEqual(edit_string_for_tags([plain, spaces]),
                         'plain, spa ces')
        self.assertEqual(edit_string_for_tags([plain, spaces, comma]),
                         'plain, spa ces, "com,ma"')
        self.assertEqual(edit_string_for_tags([plain, comma]),
                         'plain "com,ma"')
        self.assertEqual(edit_string_for_tags([comma, spaces]),
                         '"com,ma", spa ces')

    def test_tag_d_validation(self):
        t = TagField(required=False)
        self.assertEqual(t.clean(''), '')
        self.assertEqual(t.clean('foo'), 'foo')
        self.assertEqual(t.clean('foo bar baz'), 'foo bar baz')
        self.assertEqual(t.clean('foo,bar,baz'), 'foo,bar,baz')
        self.assertEqual(t.clean('foo, bar, baz'), 'foo, bar, baz')
        self.assertEqual(
            t.clean('foo qwertyuiopasdfghjklzxcvbnm'
                    'qwertyuiopasdfghjklzxcvb bar'),
            'foo qwertyuiopasdfghjklzxcvbnmqwertyuiopasdfghjklzxcvb bar')
        self.assertRaises(
            forms.ValidationError, t.clean,
            'foo qwertyuiopasdfghjklzxcvbnmqwertyuiopasdfghjklzxcvbn bar')

    def test_tag_get_from_model(self):
        FormTest.objects.create(tags='test3 test2 test1')
        FormTest.objects.create(tags='toto titi')
        self.assertEquals(FormTest.tags, 'test1 test2 test3 titi toto')


#########
# Forms #
#########


class TestTagAdminForm(TestCase):

    def test_clean_name(self):
        datas = {'name': 'tag'}
        form = TagAdminForm(datas)
        self.assertTrue(form.is_valid())

    def test_clean_name_multi(self):
        datas = {'name': 'tag error'}
        form = TagAdminForm(datas)
        self.assertFalse(form.is_valid())

    def test_clean_name_too_long(self):
        datas = {'name': 't' * (settings.MAX_TAG_LENGTH + 1)}
        form = TagAdminForm(datas)
        self.assertFalse(form.is_valid())

#########
# Views #
#########


@override_settings(
    ROOT_URLCONF='tagging.tests.urls',
    TEMPLATES=[
        {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'OPTIONS': {
                'loaders': ('tagging.tests.utils.VoidLoader',)
            }
        }
    ]
)
class TestTaggedObjectList(TestCase):

    def setUp(self):
        self.a1 = Article.objects.create(name='article 1')
        self.a2 = Article.objects.create(name='article 2')
        Tag.objects.update_tags(self.a1, 'static tag test')
        Tag.objects.update_tags(self.a2, 'static test')

    def get_view(self, url, queries=1, code=200,
                 expected_items=1,
                 friendly_context='article_list',
                 template='tests/article_list.html'):
        with self.assertNumQueries(queries):
            response = self.client.get(url)
        self.assertEquals(response.status_code, code)

        if code == 200:
            self.assertTrue(isinstance(response.context['tag'], Tag))
            self.assertEqual(len(response.context['object_list']),
                             expected_items)
            self.assertEqual(response.context['object_list'],
                             response.context[friendly_context])
            self.assertTemplateUsed(response, template)
        return response

    def test_view_static(self):
        self.get_view('/static/', expected_items=2)

    def test_view_dynamic(self):
        self.get_view('/tag/', expected_items=1)

    def test_view_related(self):
        response = self.get_view('/static/related/',
                                 queries=2, expected_items=2)
        self.assertEquals(len(response.context['related_tags']), 2)

    def test_view_no_queryset_no_model(self):
        self.assertRaises(ImproperlyConfigured, self.get_view,
                          '/no-query-no-model/')

    def test_view_no_tag(self):
        self.assertRaises(AttributeError, self.get_view, '/no-tag/')

    def test_view_404(self):
        self.get_view('/unavailable/', code=404)



"""
Admin components for tagging.
"""
from django.contrib import admin

from tagging.forms import TagAdminForm
from tagging.models import Tag
from tagging.models import TaggedItem


class TagAdmin(admin.ModelAdmin):
    form = TagAdminForm


admin.site.register(TaggedItem)
admin.site.register(Tag, TagAdmin)



"""
A custom Model Field for tagging.
"""
from django.db.models import signals
from django.db.models.fields import CharField
from django.utils.translation import gettext_lazy as _

from tagging import settings
from tagging.forms import TagField as TagFormField
from tagging.models import Tag
from tagging.utils import edit_string_for_tags


class TagField(CharField):
    """
    A "special" character field that actually works as a relationship to tags
    "under the hood". This exposes a space-separated string of tags, but does
    the splitting/reordering/etc. under the hood.
    """
    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = kwargs.get('max_length', 255)
        kwargs['blank'] = kwargs.get('blank', True)
        super(TagField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name):
        super(TagField, self).contribute_to_class(cls, name)

        # Make this object the descriptor for field access.
        setattr(cls, self.name, self)

        # Save tags back to the database post-save
        signals.post_save.connect(self._save, cls, True)

    def __get__(self, instance, owner=None):
        """
        Tag getter. Returns an instance's tags if accessed on an instance, and
        all of a model's tags if called on a class. That is, this model::

           class Link(models.Model):
               ...
               tags = TagField()

        Lets you do both of these::

           >>> l = Link.objects.get(...)
           >>> l.tags
           'tag1 tag2 tag3'

           >>> Link.tags
           'tag1 tag2 tag3 tag4'

        """
        # Handle access on the model (i.e. Link.tags)
        if instance is None:
            return edit_string_for_tags(Tag.objects.usage_for_model(owner))

        tags = self._get_instance_tag_cache(instance)
        if tags is None:
            if instance.pk is None:
                self._set_instance_tag_cache(instance, '')
            else:
                self._set_instance_tag_cache(
                    instance, edit_string_for_tags(
                        Tag.objects.get_for_object(instance)))
        return self._get_instance_tag_cache(instance)

    def __set__(self, instance, value):
        """
        Set an object's tags.
        """
        if instance is None:
            raise AttributeError(
                _('%s can only be set on instances.') % self.name)
        if settings.FORCE_LOWERCASE_TAGS and value is not None:
            value = value.lower()
        self._set_instance_tag_cache(instance, value)

    def _save(self, **kwargs):  # signal, sender, instance):
        """
        Save tags back to the database
        """
        tags = self._get_instance_tag_cache(kwargs['instance'])
        if tags is not None:
            Tag.objects.update_tags(kwargs['instance'], tags)

    def __delete__(self, instance):
        """
        Clear all of an object's tags.
        """
        self._set_instance_tag_cache(instance, '')

    def _get_instance_tag_cache(self, instance):
        """
        Helper: get an instance's tag cache.
        """
        return getattr(instance, '_%s_cache' % self.attname, None)

    def _set_instance_tag_cache(self, instance, tags):
        """
        Helper: set an instance's tag cache.
        """
        # The next instruction does nothing particular,
        # but needed to by-pass the deferred fields system
        # when saving an instance, which check the keys present
        # in instance.__dict__.
        # The issue is introducted in Django 1.10
        instance.__dict__[self.attname] = tags
        setattr(instance, '_%s_cache' % self.attname, tags)

    def get_internal_type(self):
        return 'CharField'

    def formfield(self, **kwargs):
        defaults = {'form_class': TagFormField}
        defaults.update(kwargs)
        return super(TagField, self).formfield(**defaults)



"""
Django-tagging
"""
__version__ = '0.5.3'
__license__ = 'BSD License'

__author__ = 'Jonathan Buchanan'
__author_email__ = 'jonathan.buchanan@gmail.com'

__maintainer__ = 'Arturo Inzunza'
__maintainer_email__ = 'arturo@plateiq.com'

__url__ = 'https://github.com/inzunzo/django-tagging-qubiqle'

default_app_config = 'tagging.apps.TaggingConfig'



from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False,
                    auto_created=True, primary_key=True)),
                ('name', models.CharField(
                    unique=True, max_length=50,
                    verbose_name='name', db_index=True)),
            ],
            options={
                'ordering': ('name',),
                'verbose_name': 'tag',
                'verbose_name_plural': 'tags',
            },
        ),
        migrations.CreateModel(
            name='TaggedItem',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False,
                    auto_created=True, primary_key=True)),
                ('object_id', models.PositiveIntegerField(
                    verbose_name='object id', db_index=True)),
                ('content_type', models.ForeignKey(
                    verbose_name='content type',
                    on_delete=models.SET_NULL,
                    to='contenttypes.ContentType')),
                ('tag', models.ForeignKey(
                    related_name='items', verbose_name='tag',
                    on_delete=models.SET_NULL,
                    to='tagging.Tag')),
            ],
            options={
                'verbose_name': 'tagged item',
                'verbose_name_plural': 'tagged items',
            },
        ),
        migrations.AlterUniqueTogether(
            name='taggeditem',
            unique_together=set([('tag', 'content_type', 'object_id')]),
        ),
    ]



"""
Tagging related views.
"""
from django.core.exceptions import ImproperlyConfigured
from django.http import Http404
from django.utils.translation import gettext as _
from django.views.generic.list import ListView

from tagging.models import Tag
from tagging.models import TaggedItem
from tagging.utils import get_queryset_and_model
from tagging.utils import get_tag


class TaggedObjectList(ListView):
    """
    A thin wrapper around
    ``django.views.generic.list.ListView`` which creates a
    ``QuerySet`` containing instances of the given queryset or model
    tagged with the given tag.

    In addition to the context variables set up by ``object_list``, a
    ``tag`` context variable will contain the ``Tag`` instance for the
    tag.

    If ``related_tags`` is ``True``, a ``related_tags`` context variable
    will contain tags related to the given tag for the given model.
    Additionally, if ``related_tag_counts`` is ``True``, each related
    tag will have a ``count`` attribute indicating the number of items
    which have it in addition to the given tag.
    """
    tag = None
    related_tags = False
    related_tag_counts = True

    def get_tag(self):
        if self.tag is None:
            try:
                self.tag = self.kwargs.pop('tag')
            except KeyError:
                raise AttributeError(
                    _('TaggedObjectList must be called with a tag.'))

        tag_instance = get_tag(self.tag)
        if tag_instance is None:
            raise Http404(_('No Tag found matching "%s".') % self.tag)

        return tag_instance

    def get_queryset_or_model(self):
        if self.queryset is not None:
            return self.queryset
        elif self.model is not None:
            return self.model
        else:
            raise ImproperlyConfigured(
                "%(cls)s is missing a QuerySet. Define "
                "%(cls)s.model, %(cls)s.queryset, or override "
                "%(cls)s.get_queryset_or_model()." % {
                    'cls': self.__class__.__name__
                }
            )

    def get_queryset(self):
        self.queryset_or_model = self.get_queryset_or_model()
        self.tag_instance = self.get_tag()
        return TaggedItem.objects.get_by_model(
            self.queryset_or_model, self.tag_instance)

    def get_context_data(self, **kwargs):
        context = super(TaggedObjectList, self).get_context_data(**kwargs)
        context['tag'] = self.tag_instance

        if self.related_tags:
            queryset, model = get_queryset_and_model(self.queryset_or_model)
            context['related_tags'] = Tag.objects.related_for_model(
                self.tag_instance, model, counts=self.related_tag_counts)
        return context



"""
Based entirely on Django's own ``setup.py``.
"""
from setuptools import find_packages
from setuptools import setup

import tagging

setup(
    name='django-tagging-plateiq',
    version=tagging.__version__,

    description='Generic tagging application for Django with support for CharText IDs',
    long_description='\n'.join([open('README.rst').read(),
                                open('CHANGELOG.txt').read()]),
    keywords='django, tag, tagging',

    author=tagging.__author__,
    author_email=tagging.__author_email__,
    maintainer=tagging.__maintainer__,
    maintainer_email=tagging.__maintainer_email__,
    url=tagging.__url__,
    license=tagging.__license__,

    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,

    classifiers=[
        'Framework :: Django',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
        'Topic :: Software Development :: Libraries :: Python Modules']
)



"""
Registery for tagging.
"""
from tagging.managers import ModelTaggedItemManager
from tagging.managers import TagDescriptor

registry = []


class AlreadyRegistered(Exception):
    """
    An attempt was made to register a model more than once.
    """
    pass


def register(model, tag_descriptor_attr='tags',
             tagged_item_manager_attr='tagged'):
    """
    Sets the given model class up for working with tags.
    """
    if model in registry:
        raise AlreadyRegistered(
            "The model '%s' has already been registered." %
            model._meta.object_name)
    if hasattr(model, tag_descriptor_attr):
        raise AttributeError(
            "'%s' already has an attribute '%s'. You must "
            "provide a custom tag_descriptor_attr to register." % (
                model._meta.object_name,
                tag_descriptor_attr,
            )
        )
    if hasattr(model, tagged_item_manager_attr):
        raise AttributeError(
            "'%s' already has an attribute '%s'. You must "
            "provide a custom tagged_item_manager_attr to register." % (
                model._meta.object_name,
                tagged_item_manager_attr,
            )
        )

    # Add tag descriptor
    setattr(model, tag_descriptor_attr, TagDescriptor())

    # Add custom manager
    ModelTaggedItemManager().contribute_to_class(
        model, tagged_item_manager_attr)

    # Finally register in registry
    registry.append(model)
